<?php

/**
 * Copyright 2019 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\Repository\Entities;

use JsonSerializable;

/**
 * IQRF OS and DPA entity
 */
final class OsDpa implements JsonSerializable {

	/**
	 * @var Os IQRF OS
	 */
	private $os;

	/**
	 * @var Dpa DPA
	 */
	private $dpa;

	/**
	 * @var string Notes
	 */
	private $notes;

	/**
	 * Constructor
	 * @param Os $os IQRF OS
	 * @param Dpa $dpa DPA
	 * @param string $notes Notes
	 */
	public function __construct(Os $os, Dpa $dpa, string $notes) {
		$this->os = $os;
		$this->dpa = $dpa;
		$this->notes = $notes;
	}

	/**
	 * Creates a new IQRF OS and DPA entity from an array
	 * @param array<string, string|int> $data JSON data in an array
	 * @return OsDpa Created IQRF OS and DPA entity
	 */
	public static function fromArray(array $data): self {
		$os = new Os($data['os'], $data['osVersion'], OsDpaAttributtes::fromInt($data['osAttributes']), $data['osDownloadPath']);
		$dpa = new Dpa($data['dpa'], OsDpaAttributtes::fromInt($data['dpaAttributes']), $data['dpaDownloadPath']);
		return new self($os, $dpa, $data['notes']);
	}

	/**
	 * Returns the IQRF OS build
	 * @return string IQRF OS build
	 */
	public function getOs(): string {
		return $this->os->getBuild();
	}

	/**
	 * Returns the DPA version
	 * @param bool $pretty Pretty formatting
	 * @return string DPA version
	 */
	public function getDpa(bool $pretty = false): string {
		return $this->dpa->getVersion($pretty);
	}

	/**
	 * Returns the notes
	 * @return string Notes
	 */
	public function getNotes(): string {
		return $this->notes;
	}

	/**
	 * Returns the download path
	 * @return string Download path
	 */
	public function getDownloadPath(): string {
		return $this->dpa->getDownloadPath();
	}

	/**
	 * Serialized the IQRF OS and DPA entity into JSON
	 * @return array<string, string> JSON serialized IQRF OS and DPA entity
	 */
	public function jsonSerialize(): array {
		return [
			'osBuild' => $this->os->getBuild(),
			'dpa' => $this->dpa->getVersion(true),
			'dpaRaw' => $this->dpa->getVersion(false),
			'notes' => $this->notes,
			'downloadPath' => $this->dpa->getDownloadPath(),
		];
	}

}
