<?php

/**
 * Copyright 2019 IQRF Tech s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
declare(strict_types = 1);

namespace Iqrf\Repository\Entities;

use JsonSerializable;

final class OsDpaAttributtes implements JsonSerializable {

	/**
	 * Mask for beta versions
	 */
	private const MASK_BETA = 0b1;

	/**
	 * Mask for obsolete versions
	 */
	private const MASK_OBSOLETE = 0b10;

	/**
	 * @var bool|null Beta version(s)
	 */
	private $beta;

	/**
	 * @var bool|null Obsolete version(s)
	 */
	private $obsolete;

	/**
	 * Constructor
	 * @param bool|null $beta Beta version(s)
	 * @param bool|null $obsolete Obsolete version(s)
	 */
	public function __construct(?bool $beta = null, ?bool $obsolete = null) {
		$this->beta = $beta;
		$this->obsolete = $obsolete;
	}

	/**
	 * Creates IQRF OS or DPA attributes object from an integer
	 * @param int $attributes Integer format
	 * @return static IQRF OS or DPA attributes
	 */
	public static function fromInt(int $attributes, int $mask = PHP_INT_MAX): self {
		$beta = ($mask & self::MASK_BETA) !== 0 ? ($attributes & self::MASK_BETA) !== 0 : null;
		$obsolete = ($mask & self::MASK_OBSOLETE) !== 0 ? ($attributes & self::MASK_OBSOLETE) !== 0 : null;
		return new self($beta, $obsolete);
	}

	/**
	 * Checks if the version is beta version
	 * @return bool|null Is beta version?
	 */
	public function isBeta(): ?bool {
		return $this->beta;
	}

	/**
	 * Checks if the version is obsolete version
	 * @return bool|null Is obsolete version?
	 */
	public function isObsolete(): ?bool {
		return $this->obsolete;
	}

	/**
	 * Serializes the IQRF OS or DPA attributes entity into query parameters
	 * @return array<string, string> Query parameters
	 */
	public function toQueryParameters(): array {
		$parameters = [];
		if ($this->beta !== null) {
			$parameters['beta'] = $this->beta ? 'true' : 'false';
		}
		if ($this->obsolete !== null) {
			$parameters['obsolete'] = $this->obsolete ? 'true' : 'false';
		}
		return $parameters;
	}

	/**
	 * Serializes the IQRF OS or DPA attributes entity into JSON
	 * @return array<string, bool|null> JSON serialized IQRF OS or DPA attributes entity
	 */
	public function jsonSerialize(): array {
		return [
			'beta' => $this->beta,
			'obsolete' => $this->obsolete,
		];
	}

}
